"""
This module provides aspect-oriented programming (AOP) capabilities for Python applications.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import inspect
import re
import threading
import types
from dataclasses import dataclass
from enum import auto, Enum
from typing import Optional, Dict, Type, Callable

from aspyx.di.di import order
from aspyx.reflection import Decorators, TypeDescriptor
from aspyx.di import injectable, Providers, ClassInstanceProvider, Environment, PostProcessor


class AOPException(Exception):
    """
    Exception raised for errors in the aop logic.
    """

class AspectType(Enum):
    """
    AspectType defines the types of aspect-oriented advice that can be applied to methods.

    The available types are:
    - BEFORE: Advice to be executed before the method invocation.
    - AROUND: Advice that intercepts the method invocation.
    - AFTER: Advice to be executed after the method invocation, regardless of its outcome.
    - ERROR: Advice to be executed if the method invocation raises an exception.

    These types are used to categorize and apply aspect logic at different points in a method's execution lifecycle.
    """
    BEFORE = auto()
    AROUND = auto()
    AFTER = auto()
    ERROR = auto()

class AspectTarget(ABC):
    """
    AspectTarget defines the target for an aspect. It can be used to specify the class, method, and conditions under which the aspect should be applied.
    It supports matching by class type, method name, patterns, decorators, and more.
    """

    # properties

    __slots__ = [
        "_function",
        "_type",

        "_clazz",
        "_instance",
        "names",
        "patterns",
        "types",
        "other",
        "decorators",
    ]

    # constructor

    def __init__(self):
        self._clazz = None
        self._instance = None
        self._function = None
        self._type = None

        self.patterns = []
        self.names = []
        self.types = []
        self.decorators = []

        self.other : list[AspectTarget] = []

    # abstract

    def _matches(self, clazz : Type, func):
        if not self._matches_self(clazz, func):
            for target in self.other:
                if target._matches(clazz, func):
                    return True

            return False

        return True

    @abstractmethod
    def _matches_self(self, clazz: Type, func):
        pass

    # protected

    def _add(self, target: AspectTarget):
        self.other.append(target)
        return self

     # fluent

    def function(self, func):
        self._function = func
        return self

    def type(self, type: AspectType):
        self._type = type

        return self

    def of_type(self, type: Type):
        self.types.append(type)
        return self

    def decorated_with(self, decorator):
        self.decorators.append(decorator)
        return self

    def matches(self, pattern: str):
        """
        Matches the target against a pattern.
        """
        self.patterns.append(re.compile(pattern))
        return self

    def named(self, name: str):
        self.names.append(name)
        return self

class ClassAspectTarget(AspectTarget):
    # properties

    __slots__ = [
    ]

    # public

    def _matches_self(self, clazz : Type, func):
        class_descriptor = TypeDescriptor.for_type(clazz)
        #descriptor = TypeDescriptor.for_type(func)
        # type

        if len(self.types) > 0:
            if next((type for type in self.types if issubclass(clazz, type)), None) is None:
                return False

        # decorators

        if len(self.decorators) > 0:
            if next((decorator for decorator in self.decorators if class_descriptor.has_decorator(decorator)), None) is None:
                return False

        # names

        if len(self.names) > 0:
            if next((name for name in self.names if name == clazz.__name__), None) is None:
                return False

        # patterns

        if len(self.patterns) > 0:
            if next((pattern for pattern in self.patterns if re.fullmatch(pattern, clazz.__name__) is not None), None) is None:
                return False

        return True

    # fluent

class MethodAspectTarget(AspectTarget):
    # properties

    __slots__ = [ ]

    # public

    def _matches_self(self, clazz : Type, func):
        descriptor = TypeDescriptor.for_type(clazz)

        method_descriptor = descriptor.get_method(func.__name__)

        # type

        if len(self.types) > 0:
            if next((type for type in self.types if issubclass(clazz, type)), None) is None:
                return False

        # decorators

        if len(self.decorators) > 0:
            if next((decorator for decorator in self.decorators if method_descriptor.has_decorator(decorator)), None) is None:
                return False

        # names

        if len(self.names) > 0:
            if next((name for name in self.names if name == func.__name__), None) is None:
                return False

        # patterns

        if len(self.patterns) > 0:
            if next((pattern for pattern in self.patterns if re.fullmatch(pattern, func.__name__) is not None), None) is None:
                return False

        # yipee

        return True

def methods():
    """
    Create a new AspectTarget instance to define method aspect targets.
    """
    return MethodAspectTarget()

def classes():
    """
    Create a new AspectTarget instance to define class aspect targets.
    """
    return ClassAspectTarget()


class JoinPoint:
    __slots__ = [
        "next",
    ]

    # constructor

    def __init__(self, next: 'JoinPoint'):
        self.next = next

    # public

    def call(self, invocation: 'Invocation'):
        pass

class FunctionJoinPoint(JoinPoint):
    __slots__ = [
        "instance",
        "func",
    ]

    def __init__(self, instance, func, next: Optional['JoinPoint']):
        super().__init__(next)

        self.instance = instance
        self.func = func

    def call(self, invocation: 'Invocation'):
        invocation.current_join_point = self

        return self.func(self.instance, invocation)

class MethodJoinPoint(FunctionJoinPoint):
    __slots__ = []

    def __init__(self, instance, func):
        super().__init__(instance, func, None)

    def call(self, invocation: 'Invocation'):
        invocation.current_join_point = self

        return self.func(*invocation.args, **invocation.kwargs)

@dataclass
class JoinPoints:
    before: list[JoinPoint]
    around: list[JoinPoint]
    error: list[JoinPoint]
    after: list[JoinPoint]

class Invocation:
    """
    Invocation stores the relevant data of a single method invocation.
    It holds the arguments, keyword arguments, result, error, and the join points that define the aspect behavior.
    """
    # properties

    __slots__ = [
        "func",
        "args",
        "kwargs",
        "result",
        "exception",
        "join_points",
        "current_join_point",
    ]

    # constructor

    def __init__(self, func, join_points: JoinPoints):
        self.func = func
        self.args : list[object] = []
        self.kwargs = None
        self.result = None
        self.exception = None
        self.join_points = join_points
        self.current_join_point = None

    def call(self, *args, **kwargs):
        # remember args

        self.args = args
        self.kwargs = kwargs

        # run all before

        for join_point in self.join_points.before:
            join_point.call(self)

        # run around's with the method being the last aspect!

        try:
            self.result = self.join_points.around[0].call(self) # will follow the proceed chain

        except Exception as e:
            self.exception = e
            for join_point in self.join_points.error:
                join_point.call(self)

        # run all before

        for join_point in self.join_points.after:
            join_point.call(self)

        if self.exception is not None:
            raise self.exception # rethrow the error

        return self.result

    def proceed(self, *args, **kwargs):
        """
        Proceed to the next join point in the around chain up to the original method.
        """
        if len(args) > 0 or len(kwargs) > 0:  # as soon as we have args, we replace the current ones
            self.args = args
            self.kwargs = kwargs

        # next one please...

        return self.current_join_point.next.call(self)

@injectable()
class Advice:
    # static data

    targets: list[AspectTarget] = []

    __slots__ = [
        "cache",
        "lock"
    ]

    # constructor

    def __init__(self):
        self.cache : Dict[Type, Dict[Callable,JoinPoints]] = {}
        self.lock = threading.RLock()

    # methods

    def collect(self, clazz, member, type: AspectType, environment: Environment):
        aspects = [FunctionJoinPoint(environment.get(target._clazz), target._function, None) for target in Advice.targets if target._type == type and target._matches(clazz, member)]

        # link

        for i in range(0, len(aspects) - 1):
            aspects[i].next = aspects[i + 1]

        # done

        return aspects

    def join_points4(self, instance, environment: Environment) -> Dict[Callable,JoinPoints]:
        clazz = type(instance)

        result = self.cache.get(clazz, None)
        if result is None:
            with self.lock:
                result = self.cache.get(clazz, None)

                if result is None:
                    result = {}

                    for _, member in inspect.getmembers(clazz, predicate=inspect.isfunction):
                        join_points = self.compute_join_points(clazz, member, environment)
                        if join_points is not None:
                            result[member] = join_points

                    self.cache[clazz] = result

        # add around methods

        value = {}

        for key, cjp in result.items():
            jp = JoinPoints(
                before=cjp.before,
                around=cjp.around,
                error=cjp.error,
                after=cjp.after)

            # add method to around

            jp.around.append(MethodJoinPoint(instance, key))
            if len(jp.around) > 1:
                jp.around[len(jp.around) - 2].next = jp.around[len(jp.around) - 1]

            value[key] = jp

        # done

        return value

    def compute_join_points(self, clazz, member, environment: Environment) -> Optional[JoinPoints]:
        befores = self.collect(clazz, member, AspectType.BEFORE, environment)
        arounds = self.collect(clazz, member, AspectType.AROUND, environment)
        afters = self.collect(clazz, member, AspectType.AFTER, environment)
        errors = self.collect(clazz, member, AspectType.ERROR, environment)

        if len(befores) > 0 or len(arounds) > 0 or len(afters) > 0  or len(errors) > 0:
            return JoinPoints(
                before=befores,
                around=arounds,
                error=errors,
                after=afters
            )
        else:
            return None

def sanity_check(clazz: Type, name: str):
    m = TypeDescriptor.for_type(clazz).get_method(name)
    if len(m.param_types) != 1 or m.param_types[0] != Invocation:
        raise AOPException(f"Method {clazz.__name__}.{name} expected to have one parameter of type Invocation")

# decorators

def advice(cls):
    """
    Classes decorated with @advice are treated as aspect classes.
    They can contain methods decorated with @before, @after, @around, or @error to define aspects.
    """
    Providers.register(ClassInstanceProvider(cls, True))

    Decorators.add(cls, advice)

    for name, member in TypeDescriptor.for_type(cls).methods.items():
        decorator = next((decorator for decorator in member.decorators if decorator.decorator in [before, after, around, error]), None)
        if decorator is not None:
            target = decorator.args[0]
            target._clazz = cls
            sanity_check(cls, name)
            Advice.targets.append(target)

    return cls


# decorators

def _register(decorator, targets: list[AspectTarget], func, aspect_type: AspectType):
    target = targets[0]

    for i in range(1, len(targets)):
        target._add(targets[i])

    target.function(func).type(aspect_type)

    Decorators.add(func, decorator, target)

def before(*targets: AspectTarget):
    """
    Methods decorated with @before will be executed before the target method is invoked.
    """
    def decorator(func):
        _register(before, targets, func, AspectType.BEFORE)

        return func

    return decorator

def error(*targets: AspectTarget):
    """
    Methods decorated with @error will be executed if the target method raises an exception."""
    def decorator(func):
        _register(error, targets, func, AspectType.ERROR)

        return func

    return decorator

def after(*targets: AspectTarget):
    """
    Methods decorated with @after will be executed after the target method is invoked.
    """
    def decorator(func):
        _register(after, targets, func, AspectType.AFTER)

        return func

    return decorator

def around(*targets: AspectTarget):
    """
    Methods decorated with @around will be executed around the target method.
    Every around method must accept a single parameter of type Invocation and needs to call proceed
    on this parameter to proceed to the next around method.
    """
    def decorator(func):
        _register(around, targets, func, AspectType.AROUND)

        return func

    return decorator

@injectable()
@order(0)
class AdviceProcessor(PostProcessor):
    # properties

    __slots__ = [
        "advice",
    ]

    # constructor

    def __init__(self, advice: Advice):
        super().__init__()

        self.advice = advice

    # implement

    def process(self, instance: object, environment: Environment):
        join_point_dict = self.advice.join_points4(instance, environment)

        for member, join_points in join_point_dict.items():
            Environment.logger.debug("add aspects for %s:%s", type(instance), member.__name__)

            def wrap(jp):
                return lambda *args, **kwargs: Invocation(member, jp).call(*args, **kwargs)

            setattr(instance, member.__name__, types.MethodType(wrap(join_points), instance))



"""
Configuration handling module.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import os
from typing import Optional, Type, TypeVar
from dotenv import load_dotenv

from aspyx.di import injectable, Environment, LifecycleCallable, Lifecycle
from aspyx.di.di import order, inject
from aspyx.reflection import Decorators, DecoratorDescriptor, TypeDescriptor

T = TypeVar("T")

class ConfigurationException(Exception):
    """
    Exception raised for errors in the configuration logic.
    """

@injectable()
class ConfigurationManager:
    """
    ConfigurationManager is responsible for managing different configuration sources by merging the different values
    and offering a uniform api.
    """

    __slots__ = ["sources", "_data", "coercions"]

    # constructor

    def __init__(self):
        self.sources = []
        self._data = {}
        self.coercions = {
            int: int,
            float: float,
            bool: lambda v: str(v).lower() in ("1", "true", "yes", "on"),
            str: str,
            # Add more types as needed
        }

    # internal

    def _register(self, source: ConfigurationSource):
        self.sources.append(source)
        self.load_source(source)

    # public

    def load_source(self,  source: ConfigurationSource):
        def merge_dicts(a: dict, b: dict) -> dict:
            result = a.copy()
            for key, b_val in b.items():
                if key in result:
                    a_val = result[key]
                    if isinstance(a_val, dict) and isinstance(b_val, dict):
                        result[key] = merge_dicts(a_val, b_val)  # Recurse
                    else:
                        result[key] = b_val  # Overwrite
                else:
                    result[key] = b_val
            return result

        self._data = merge_dicts(self._data, source.load())

    def get(self, path: str, type: Type[T], default : Optional[T]=None) -> T:
        """
        Get a configuration value by path and type, with optional coercion.
        Arguments:
            path (str): The path to the configuration value, e.g. "database.host".
            type (Type[T]): The expected type.
            default (Optional[T]): The default value to return if the path is not found.
        Returns:
            T: The configuration value coerced to the specified type, or the default value if not found.
        """
        def resolve_value(path: str, default=None) -> T:
            keys = path.split(".")
            current = self._data
            for key in keys:
                if not isinstance(current, dict) or key not in current:
                    return default
                current = current[key]

            return current

        v = resolve_value(path, default)

        if isinstance(v, type):
            return v

        if type in self.coercions:
            try:
                return self.coercions[type](v)
            except Exception as e:
                raise ConfigurationException(f"error during coercion to {type}") from e
        else:
            raise ConfigurationException(f"unknown coercion to {type}")


class ConfigurationSource(ABC):
    """
    A configuration source is a provider of configuration data.
    """

    __slots__ = []

    def __init__(self):
        pass

    @inject()
    def set_manager(self, manager: ConfigurationManager):
        manager._register(self)

    @abstractmethod
    def load(self) -> dict:
        """
        return the configuration values of this source as a dictionary.
        """

@injectable()
class EnvConfigurationSource(ConfigurationSource):
    """
    EnvConfigurationSource loads all environment variables.
    """

    __slots__ = []

    # constructor

    def __init__(self):
        super().__init__()

        load_dotenv()

    # implement

    def load(self) -> dict:
        def merge_dicts(a, b):
            """Recursively merges b into a"""
            for key, value in b.items():
                if isinstance(value, dict) and key in a and isinstance(a[key], dict):
                    merge_dicts(a[key], value)
                else:
                    a[key] = value
            return a

        def explode_key(key, value):
            """Explodes keys with '.' or '/' into nested dictionaries"""
            parts = key.replace('/', '.').split('.')
            d = current = {}
            for part in parts[:-1]:
                current[part] = {}
                current = current[part]
            current[parts[-1]] = value
            return d

        exploded = {}

        for key, value in os.environ.items():
            if '.' in key or '/' in key:
                partial = explode_key(key, value)
                merge_dicts(exploded, partial)
            else:
                exploded[key] = value

        return exploded

# decorator

def value(key: str, default=None):
    """
    Decorator to inject a configuration value into a method.

    Arguments:
        key (str): The configuration key to inject.
        default: The default value to use if the key is not found.

    """
    def decorator(func):
        Decorators.add(func, value, key, default)

        return func

    return decorator

@injectable()
@order(9)
class ConfigurationLifecycleCallable(LifecycleCallable):
    def __init__(self,  manager: ConfigurationManager):
        super().__init__(value, Lifecycle.ON_INJECT)

        self.manager = manager

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [self.manager.get(decorator.args[0], method.param_types[0], decorator.args[1])]



"""
Test cases for the aspyx.di module.
"""
from __future__ import annotations

import time
import logging
import unittest
from typing import Dict

from aspyx.di import InjectorException, injectable, on_init, on_running, on_destroy, inject_environment, inject, Factory, create, environment, Environment, PostProcessor, factory
from aspyx.di.di import order, on_running

from .di_import import ImportedEnvironment

# not here

logging.basicConfig(
    level=logging.INFO,
    format='[%(asctime)s] %(levelname)s in %(filename)s:%(lineno)d - %(message)s'
)

def configure_logging(levels: Dict[str, int]) -> None:
    for name in levels:
        logging.getLogger(name).setLevel(levels[name])

#configure_logging({"aspyx": logging.DEBUG})

# not here


@injectable()
@order(10)
class SamplePostProcessor(PostProcessor):
    def process(self, instance: object, environment: Environment):
        pass #print(f"created a {instance}")

class Foo:
    def __init__(self):
        self.inited = False

    @on_init()
    def init(self):
        self.inited = True

class Baz:
    def __init__(self):
        self.inited = False

    @on_init()
    def init(self):
        self.inited = True

@injectable()
class Bazong:
    def __init__(self):
        pass

class Base:
    def __init__(self):
        pass

class Ambiguous:
    def __init__(self):
        pass

class Unknown:
    def __init__(self):
        pass#

@injectable(scope="request")
class NonSingleton:
    def __init__(self):
        super().__init__()

@injectable()
class Derived(Ambiguous):
    def __init__(self):
        super().__init__()

@injectable()
class Derived1(Ambiguous):
    def __init__(self):
        super().__init__()

@injectable()
class Bar(Base):
    def __init__(self, foo: Foo):
        super().__init__()

        self.bazong = None
        self.baz = None
        self.foo = foo
        self.inited = False
        self.running = False
        self.destroyed = False
        self.environment = None

    @on_init()
    def init(self):
        self.inited = True

    @on_running()
    def running(self):
        self.running = True

    @on_destroy()
    def destroy(self):
        self.destroyed = True

    @inject_environment()
    def init_environment(self, env: Environment):
        self.environment = env

    @inject()
    def set(self, baz: Baz, bazong: Bazong) -> None:
        self.baz = baz
        self.bazong = bazong

@factory()
class SampleFactory(Factory[Foo]):
    __slots__ = []

    def __init__(self):
        pass

    def create(self) -> Foo:
        return Foo()

@environment()
class SimpleEnvironment:
    # constructor

    def __init__(self):
        pass

    # create some beans

    @create()
    def create(self) -> Baz:
        return Baz()

@environment(imports=[SimpleEnvironment, ImportedEnvironment])
class ComplexEnvironment:
    # constructor

    def __init__(self):
        pass

class TestDI(unittest.TestCase):
    testEnvironment = Environment(SimpleEnvironment)

    def test_process_factory_instances(self):
        env = TestDI.testEnvironment

        print(env.report())

        baz = env.get(Baz)
        foo = env.get(Foo)
        self.assertEqual(baz.inited, True)
        self.assertEqual(foo.inited, True)

    def test_inject_base_class(self):
        env = TestDI.testEnvironment

        base = env.get(Base)
        self.assertEqual(type(base), Bar)

    def test_inject_ambiguous_class(self):
        with self.assertRaises(InjectorException):
            env = TestDI.testEnvironment
            env.get(Ambiguous)

    def test_create_unknown(self):
        with self.assertRaises(InjectorException):
            env = TestDI.testEnvironment
            env.get(Unknown)

    def test_inject_constructor(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)
        baz = env.get(Baz)
        bazong = env.get(Bazong)
        foo = env.get(Foo)

        self.assertIsNotNone(bar)
        self.assertIs(bar.foo, foo)
        self.assertIs(bar.baz, baz)
        self.assertIs(bar.bazong, bazong)

    def test_factory(self):
        env = TestDI.testEnvironment
        foo = env.get(Foo)
        self.assertIsNotNone(foo)

    def test_create_factory(self):
        env = TestDI.testEnvironment
        baz = env.get(Baz)
        self.assertIsNotNone(baz)

    def test_singleton(self):
        env = TestDI.testEnvironment

        # injectable

        bar = env.get(Bar)
        bar1 = env.get(Bar)
        self.assertIs(bar, bar1)

        # factory

        foo = env.get(Foo)
        foo1 = env.get(Foo)
        self.assertIs(foo,foo1)

        # create

        baz  = env.get(Baz)
        baz1 = env.get(Baz)
        self.assertIs(baz, baz1)

    def test_non_singleton(self):
        env = TestDI.testEnvironment

        ns = env.get(NonSingleton)
        ns1 = env.get(NonSingleton)

        self.assertIsNot(ns, ns1)

    #def test_import_configurations(self): TODO
    #    env = Environment(TestEnvironment)#

    #    imported = env.get(ImportedClass)

    #    self.assertIsNotNone(imported)

    def test_init(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        self.assertEqual(bar.inited, True)

    def test_running(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        self.assertEqual(bar.running, True)

    def test_destroy(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        env.destroy()

        self.assertEqual(bar.destroyed, True)

    def test_performance(self):
        env = TestDI.testEnvironment

        start = time.perf_counter()
        for _ in range(1000000):
            env.get(Bar)

        end = time.perf_counter()

        avg_ms = ((end - start) / 1000000) * 1000
        print(f"Average time per Bar creation: {avg_ms:.3f} ms")


if __name__ == '__main__':
    unittest.main()



"""
The deoendency injection module provides a framework for managing dependencies and lifecycle of objects in Python applications.
"""
from __future__ import annotations

import inspect
import logging

from abc import abstractmethod, ABC
from enum import Enum, auto
import threading
from typing import Type, Dict, TypeVar, Generic, Optional, cast, Callable

from aspyx.di.util import StringBuilder
from aspyx.reflection import Decorators, TypeDescriptor, DecoratorDescriptor

T = TypeVar("T")

class Factory(ABC, Generic[T]):
    """
    Abstract base class for factories that create instances of type T.
    """

    __slots__ = []

    @abstractmethod
    def create(self) -> T:
        pass

class InjectorException(Exception):
    """
    Exception raised for errors in the injector.
    """

class AbstractInstanceProvider(ABC, Generic[T]):
    """
    Interface for instance providers.
    """
    @abstractmethod
    def get_module(self) -> str:
        pass

    @abstractmethod
    def get_type(self) -> Type[T]:
        pass

    @abstractmethod
    def is_eager(self) -> bool:
        pass

    @abstractmethod
    def get_scope(self) -> str:
        pass

    @abstractmethod
    def get_dependencies(self) -> list[AbstractInstanceProvider]:
        pass

    @abstractmethod
    def create(self, environment: Environment, *args):
        pass

    @abstractmethod
    def resolve(self, context: Providers.Context):
        pass


class InstanceProvider(AbstractInstanceProvider):
    """
    An InstanceProvider is able to create instances of type T.
    """
    __slots__ = [
        "host",
        "type",
        "eager",
        "scope",
        "dependencies"
    ]

    # constructor

    def __init__(self, host: Type, t: Type[T], eager: bool, scope: str):
        self.host = host
        self.type = t
        self.eager = eager
        self.scope = scope
        self.dependencies : Optional[list[AbstractInstanceProvider]] = None

    # internal

    def _is_resolved(self) -> bool:
        return self.dependencies is not None

    # implement AbstractInstanceProvider

    def resolve(self, context: Providers.Context):
        pass

    def get_module(self) -> str:
        return self.host.__module__

    def get_type(self) -> Type[T]:
        return self.type

    def is_eager(self) -> bool:
        return self.eager

    def get_scope(self) -> str:
        return self.scope

    def get_dependencies(self) -> list[AbstractInstanceProvider]:
        return self.dependencies

    # public

    def module(self) -> str:
        return self.host.__module__

    def add_dependency(self, provider: AbstractInstanceProvider):
        if any(issubclass(provider.get_type(), dependency.get_type()) for dependency in self.dependencies):
            return False

        self.dependencies.append(provider)

        return True

    @abstractmethod
    def create(self, environment: Environment, *args):
        pass

# we need this classes to bootstrap the system...
class SingletonScopeInstanceProvider(InstanceProvider):
    def __init__(self):
        super().__init__(SingletonScopeInstanceProvider, SingletonScope, False, "request")

    def create(self, environment: Environment, *args):
        return SingletonScope()

class RequestScopeInstanceProvider(InstanceProvider):
    def __init__(self):
        super().__init__(RequestScopeInstanceProvider, RequestScope, False, "singleton")

    def create(self, environment: Environment, *args):
        return RequestScope()


class AmbiguousProvider(AbstractInstanceProvider):
    """
    An AmbiguousProvider covers all cases, where fetching a class would lead to an ambiguity exception.
    """

    __slots__ = [
        "type",
        "providers",
    ]

    # constructor

    def __init__(self, type: Type, *providers: AbstractInstanceProvider):
        super().__init__()

        self.type = type
        self.providers = list(providers)

    # public

    def add_provider(self, provider: AbstractInstanceProvider):
        self.providers.append(provider)

    # implement

    def get_module(self) -> str:
        return self.type.__module__

    def get_type(self) -> Type[T]:
        return self.type

    def is_eager(self) -> bool:
        return False

    def get_scope(self) -> str:
        return "singleton"

    def get_dependencies(self) -> list[AbstractInstanceProvider]:
        return []

    def resolve(self, context: Providers.Context):
        pass

    def create(self, environment: Environment, *args):
        raise InjectorException(f"multiple candidates for type {self.type}")

    def __str__(self):
        return f"AmbiguousProvider({self.type})"

class Scopes:
    # static data

    scopes : Dict[str, Type] = {}

    # class methods

    @classmethod
    def get(cls, scope: str, environment: Environment):
        scope_type = Scopes.scopes.get(scope, None)
        if scope_type is None:
            raise InjectorException(f"unknown scope type {scope}")

        return environment.get(scope_type)

    @classmethod
    def register(cls, scope_type: Type, name: str):
        Scopes.scopes[name] = scope_type

class Scope:
    # properties

    __slots__ = [
    ]

    # constructor

    def __init__(self):
        pass

    # public

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        return provider.create(environment, *arg_provider())

class EnvironmentInstanceProvider(AbstractInstanceProvider):
    # properties

    __slots__ = [
        "environment",
        "scope_instance",
        "provider",
        "dependencies",
    ]

    # constructor

    def __init__(self, environment: Environment, provider: AbstractInstanceProvider):
        super().__init__()

        self.environment = environment
        self.provider = provider
        self.dependencies : list[AbstractInstanceProvider] = []

        self.scope_instance = Scopes.get(provider.get_scope(), environment)

    # implement

    def resolve(self, context: Providers.Context):
        pass # noop

    def get_module(self) -> str:
        return self.provider.get_module()

    def get_type(self) -> Type[T]:
        return self.provider.get_type()

    def is_eager(self) -> bool:
        return self.provider.is_eager()

    def get_scope(self) -> str:
        return self.provider.get_scope()

    # custom logic

    def tweak_dependencies(self, providers: dict[Type, AbstractInstanceProvider]):
        for dependency in self.provider.get_dependencies():
            instance_provider = providers.get(dependency.get_type(), None)
            if instance_provider is None:
                raise InjectorException(f"missing import for {dependency.get_type()} ")

            self.dependencies.append(instance_provider)

    def get_dependencies(self) -> list[AbstractInstanceProvider]:
        return self.provider.get_dependencies()

    def create(self, environment: Environment, *args):
        return self.scope_instance.get(self.provider, self.environment, lambda: [provider.create(environment) for provider in self.dependencies]) # already scope property!

    def __str__(self):
        return f"EnvironmentInstanceProvider({self.provider})"

class ClassInstanceProvider(InstanceProvider):
    """
    A ClassInstanceProvider is able to create instances of type T by calling the class constructor.
    """

    __slots__ = [
        "params"
    ]

    # constructor

    def __init__(self, t: Type[T], eager: bool, scope = "singleton"):
        super().__init__(t, t, eager, scope)

        self.params = 0

    # implement

    def resolve(self, context: Providers.Context):
        context.add(self)

        if not self._is_resolved():
            self.dependencies = []

            # check constructor

            init = TypeDescriptor.for_type(self.type).get_method("__init__")
            if init is None:
                raise InjectorException(f"{self.type.__name__} does not implement __init__")

            for param in init.param_types:
                provider = Providers.get_provider(param)
                self.params += 1
                if self.add_dependency(provider): # a dependency can occur multiple times, e.g in __init__ and in an injected method
                    provider.resolve(context)

            # check @inject

            for method in TypeDescriptor.for_type(self.type).get_methods():
                if method.has_decorator(inject):
                    for param in method.param_types:
                        provider = Providers.get_provider(param)

                        if self.add_dependency(provider):
                            provider.resolve(context)
        else: # check if the dependencies create a cycle
            context.add(*self.dependencies)

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        return environment.created(self.type(*args[:self.params]))

    # object

    def __str__(self):
        return f"ClassInstanceProvider({self.type.__name__})"

class FunctionInstanceProvider(InstanceProvider):
    """
    A FunctionInstanceProvider is able to create instances of type T by calling specific methods annotated with 'create".
    """

    __slots__ = [
        "method"
    ]

    # constructor

    def __init__(self, clazz : Type, method, return_type : Type[T], eager = True, scope = "singleton"):
        super().__init__(clazz, return_type, eager, scope)

        self.method = method

    # implement

    def resolve(self, context: Providers.Context):
        context.add(self)

        if not self._is_resolved():
            self.dependencies = []

            provider = Providers.get_provider(self.host)
            if self.add_dependency(provider):
                provider.resolve(context)
        else: # check if the dependencies crate a cycle
            context.add(*self.dependencies)

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        instance = self.method(*args)

        return environment.created(instance)

    def __str__(self):
        return f"FunctionInstanceProvider({self.host.__name__}.{self.method.__name__} -> {self.type.__name__})"

class FactoryInstanceProvider(InstanceProvider):
    """
    A FactoryInstanceProvider is able to create instances of type T by calling registered Factory instances.
    """

    __slots__ = []

    # class method

    @classmethod
    def get_factory_type(cls, clazz):
        return TypeDescriptor.for_type(clazz).get_method("create", local=True).return_type

    # constructor

    def __init__(self, factory: Type, eager: bool, scope: str):
        super().__init__(factory, FactoryInstanceProvider.get_factory_type(factory), eager, scope)

    # implement

    def resolve(self, context: Providers.Context):
        context.add(self)

        if  not self._is_resolved():
            self.dependencies = []

            provider = Providers.get_provider(self.host)
            if self.add_dependency(provider):
                provider.resolve(context)

        else: # check if the dependencies crate a cycle
            context.add(*self.dependencies)

        return self

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        return environment.created(args[0].create())

    def __str__(self):
        return f"FactoryInstanceProvider({self.host.__name__} -> {self.type.__name__})"


class Lifecycle(Enum):
    """
    This enum defines the lifecycle events that can be processed by lifecycle processors.
    """

    __slots__ = []

    ON_INJECT  = 0
    ON_INIT    = 1
    ON_RUNNING = 2
    ON_DESTROY = 3

class LifecycleProcessor(ABC):
    """
    A LifecycleProcessor is used to perform any side effects on managed objects during their lifecycle.
    """
    __slots__ = [
        "order"
    ]

    # constructor

    def __init__(self):
        self.order = 0
        if TypeDescriptor.for_type(type(self)).has_decorator(order):
            self.order =  TypeDescriptor.for_type(type(self)).get_decorator(order).args[0]

    # methods

    @abstractmethod
    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        pass

class PostProcessor(LifecycleProcessor):
    """
    Base class for custom post processors that are executed after object creation.
    """
    __slots__ = []


    def process(self, instance: object, environment: Environment):
        pass

    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        if lifecycle == Lifecycle.ON_INIT:
            self.process(instance, environment)


class Providers:
    """
    The Providers class is a static class that manages the registration and resolution of InstanceProviders.
    """
    # local class

    class Context:
        __slots__ = ["dependencies"]

        def __init__(self):
            self.dependencies : list[AbstractInstanceProvider] = []

        def add(self, *providers: AbstractInstanceProvider):
            for provider in providers:
                if next((p for p in self.dependencies if p.get_type() is provider.get_type()), None) is not None:
                    raise InjectorException(self.cycle_report(provider))

                self.dependencies.append(provider)

        def cycle_report(self, provider: AbstractInstanceProvider):
            cycle = ""

            first = True
            for p in self.dependencies:
                if not first:
                    cycle += " -> "

                first = False

                cycle += f"{p.get_type().__name__}"

            cycle += f" <> {provider.get_type().__name__}"

            return cycle


    # class properties

    check: list[AbstractInstanceProvider] = []

    providers : Dict[Type,AbstractInstanceProvider] = {}
    cache: Dict[Type, AbstractInstanceProvider] = {}

    resolved = False

    @classmethod
    def register(cls, provider: AbstractInstanceProvider):
        Environment.logger.debug("register provider %s(%s)", provider.get_type().__qualname__, provider.get_type().__name__)

        # local functions

        def is_injectable(type: Type) -> bool:
            if type is object:
                return False

            if inspect.isabstract(type):
                return False

            #for decorator in Decorators.get(type):
            #    if decorator.decorator is injectable:
            #        return True

            # darn

            return True

        def cache_provider_for_type(provider: AbstractInstanceProvider, type: Type):
            existing_provider = Providers.cache.get(type)
            if existing_provider is None:
                Providers.cache[type] = provider

            else:
                if type is provider.get_type():
                    raise InjectorException(f"{type} already registered")

                if isinstance(existing_provider, AmbiguousProvider):
                    cast(AmbiguousProvider, existing_provider).add_provider(provider)
                else:
                    Providers.cache[type] = AmbiguousProvider(type, existing_provider, provider)

            # recursion

            for super_class in type.__bases__:
                if is_injectable(super_class):
                    cache_provider_for_type(provider, super_class)

        # go

        Providers.check.append(provider)

        Providers.providers[provider.get_type()] = provider

        # cache providers

        cache_provider_for_type(provider, provider.get_type())

    @classmethod
    def resolve(cls):
        for provider in Providers.check:
            provider.resolve(Providers.Context())

        Providers.check.clear()

    @classmethod
    def report(cls):
        for provider in Providers.cache.values():
            print(f"provider {provider.get_type().__qualname__}")

    @classmethod
    def get_provider(cls, type: Type) -> AbstractInstanceProvider:
        provider = Providers.cache.get(type, None)
        if provider is None:
            raise InjectorException(f"{type.__name__} not registered as injectable")

        return provider

def register_factories(cls: Type):
    descriptor = TypeDescriptor.for_type(cls)

    for method in descriptor.get_methods():
        if method.has_decorator(create):
            create_decorator = method.get_decorator(create)
            Providers.register(FunctionInstanceProvider(cls, method.method, method.return_type, create_decorator.args[0],
                                                        create_decorator.args[1]))
def order(prio = 0):
    def decorator(cls):
        Decorators.add(cls, order, prio)

        return cls

    return decorator

def injectable(eager=True, scope="singleton"):
    """
    Instances of classes that are annotated with @injectable can be created by an Environment.
    """
    def decorator(cls):
        Decorators.add(cls, injectable)

        Providers.register(ClassInstanceProvider(cls, eager, scope))

        return cls

    return decorator

def factory(eager=True, scope="singleton"):
    """
    Decorator that needs to be used on a class that implements the Factory interface.
    """
    def decorator(cls):
        Decorators.add(cls, factory)

        Providers.register(ClassInstanceProvider(cls, eager, scope))
        Providers.register(FactoryInstanceProvider(cls, eager, scope))

        return cls

    return decorator

def create(eager=True, scope="singleton"):
    """
    Any method annotated with @create will be registered as a factory method.
    """
    def decorator(func):
        Decorators.add(func, create, eager, scope)
        return func

    return decorator

def on_init():
    """
    Methods annotated with @on_init will be called when the instance is created."""
    def decorator(func):
        Decorators.add(func, on_init)
        return func

    return decorator

def on_running():
    """
    Methods annotated with @on_running will be called when the container up and running."""
    def decorator(func):
        Decorators.add(func, on_running)
        return func

    return decorator

def on_destroy():
    """
    Methods annotated with @on_destroy will be called when the instance is destroyed.
    """
    def decorator(func):
        Decorators.add(func, on_destroy)
        return func

    return decorator

def environment(imports: Optional[list[Type]] = None):
    """
    This annotation is used to mark classes that control the set of injectables that will be managed based on their location
    relative to the module of the class. All @injectable s and @factory s that are located in the same or any sub-module will
    be registered and managed accordingly.
    Arguments:
        imports (Optional[list[Type]]): Optional list of imported environment types
    """
    def decorator(cls):
        Providers.register(ClassInstanceProvider(cls, True))

        Decorators.add(cls, environment, imports)
        Decorators.add(cls, injectable) # do we need that?

        register_factories(cls)

        return cls

    return decorator

def inject():
    """
    Methods annotated with @inject will be called with the required dependencies injected.
    """
    def decorator(func):
        Decorators.add(func, inject)
        return func

    return decorator

def inject_environment():
    """
    Methods annotated with @inject_environment will be called with the Environment instance injected.
    """
    def decorator(func):
        Decorators.add(func, inject_environment)
        return func

    return decorator

class Environment:
    """
    Central class that manages the lifecycle of instances and their dependencies.
    """

    # static data

    logger = logging.getLogger(__name__)  # __name__ = module name

    instance : 'Environment' = None

    __slots__ = [
        "type",
        "providers",
        "lifecycle_processors",
        "parent",
        "instances"
    ]

    # constructor

    def __init__(self, env: Type, parent : Optional[Environment] = None):
        """
        Creates a new Environment instance.

        Args:
            env (Type): The environment class that controls the scanning of managed objects.
            parent (Optional[Environment]): Optional parent environment, whose objects are inherited.
        """
        # initialize

        self.type = env
        self.parent = parent
        if self.parent is None and env is not BootEnvironment:
            self.parent = BootEnvironment.get_instance() # inherit environment including its manged instances!

        self.providers: Dict[Type, AbstractInstanceProvider] = {}
        self.lifecycle_processors: list[LifecycleProcessor] = []

        if self.parent is not None:
            self.providers |= self.parent.providers
            self.lifecycle_processors += self.parent.lifecycle_processors

        self.instances = []

        Environment.instance = self

        # resolve providers on a static basis. This is executed for all new providers ( in case of new modules multiple times) !

        Providers.resolve()

        loaded = set()

        def add_provider(type: Type, provider: AbstractInstanceProvider):
            Environment.logger.debug("\tadd provider %s for %s", provider, type)

            self.providers[type] = provider

        # bootstrapping hack, they will be overwritten by the "real" providers

        if env is BootEnvironment:
            add_provider(SingletonScope, SingletonScopeInstanceProvider())
            add_provider(RequestScope, RequestScopeInstanceProvider())

        def load_environment(env: Type):
            if env not in loaded:
                Environment.logger.debug("load environment %s", env.__qualname__)

                loaded.add(env)

                # sanity check

                decorator = TypeDescriptor.for_type(env).get_decorator(environment)
                if decorator is None:
                    raise InjectorException(f"{env.__name__} is not an environment class")

                scan = env.__module__
                if "." in scan:
                    scan = scan.rsplit('.', 1)[0]

                # recursion

                for import_environment in decorator.args[0] or []:
                    load_environment(import_environment)

                # load providers

                local_providers = {type: provider for type, provider in Providers.cache.items() if provider.get_module().startswith(scan)}

                # register providers

                # make sure, that for every type only a single EnvironmentInstanceProvider is created!
                # otherwise inheritance will fuck it up

                environment_providers : dict[AbstractInstanceProvider, EnvironmentInstanceProvider] = {}

                for type, provider in local_providers.items():
                    environment_provider = environment_providers.get(provider, None)
                    if environment_provider is None:
                        environment_provider =  EnvironmentInstanceProvider(self, provider)
                        environment_providers[provider] = environment_provider

                    add_provider(type, environment_provider)

                # tweak dependencies

                for type, provider in local_providers.items():
                    cast(EnvironmentInstanceProvider, self.providers[type]).tweak_dependencies(self.providers)

                # return local providers

                return environment_providers.values()
            else:
                return []

        # construct eager objects for local providers

        for provider in load_environment(env):
            if provider.is_eager():
                provider.create(self)

        # running callback

        for instance in self.instances:
            self.execute_processors(Lifecycle.ON_RUNNING, instance)

    # internal

    def execute_processors(self, lifecycle: Lifecycle, instance: T) -> T:
        for processor in self.lifecycle_processors:
            processor.process_lifecycle(lifecycle, instance, self)

        return instance

    def created(self, instance: T) -> T:
        # remember lifecycle processors

        if isinstance(instance, LifecycleProcessor):
            self.lifecycle_processors.append(instance)

            # sort immediately

            self.lifecycle_processors.sort(key=lambda processor: processor.order)

        # remember instance

        self.instances.append(instance)

        # execute processors

        self.execute_processors(Lifecycle.ON_INJECT, instance)
        self.execute_processors(Lifecycle.ON_INIT, instance)

        return instance

    # public

    def report(self):
        builder = StringBuilder()

        builder.append(f"Environment {self.type.__name__}")

        if self.parent is not None:
            builder.append(f" parent {self.parent.type.__name__}")

        builder.append("\n")

        # post processors

        builder.append("Processors \n")
        for processor in self.lifecycle_processors:
            builder.append(f"- {processor.__class__.__name__}\n")

        # providers

        builder.append("Providers \n")
        for result_type, provider in self.providers.items():
            if cast(EnvironmentInstanceProvider, provider).environment is self:
                builder.append(f"- {result_type.__name__}: {cast(EnvironmentInstanceProvider, provider).provider}\n")

        # instances

        builder.append("Instances \n")

        result = {}
        for obj in self.instances:
            cls = type(obj)
            result[cls] = result.get(cls, 0) + 1

        for cls, count in result.items():
            builder.append(f"- {cls.__name__}: {count} \n")

        # done

        result = str(builder)

        return result

    def destroy(self):
        """
        destroy all managed instances by calling the appropriate lifecycle methods
        """
        for instance in self.instances:
            self.execute_processors(Lifecycle.ON_DESTROY, instance)

        self.instances.clear() # make the cy happy

    def get(self, type: Type[T]) -> T:
        """
        Return and possibly create a new instance of the given type.

        Arguments:
            type (Type): The desired type

        Returns: The requested instance
        """
        provider = self.providers.get(type, None)
        if provider is None:
            Environment.logger.error("%s is not supported", type)
            raise InjectorException(f"{type} is not supported")

        return provider.create(self)

class LifecycleCallable:
    __slots__ = [
        "decorator",
        "lifecycle",
        "order"
    ]

    def __init__(self, decorator, lifecycle: Lifecycle):
        self.decorator = decorator
        self.lifecycle = lifecycle
        self.order = 0

        if TypeDescriptor.for_type(type(self)).has_decorator(order):
            self.order = TypeDescriptor.for_type(type(self)).get_decorator(order).args[0]

        AbstractCallableProcessor.register(self)

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return []


class AbstractCallableProcessor(LifecycleProcessor):
    # local classes

    class MethodCall:
        __slots__ = [
            "decorator",
            "method",
            "lifecycle_callable"
        ]

        # constructor

        def __init__(self, method: TypeDescriptor.MethodDescriptor, decorator: DecoratorDescriptor, lifecycle_callable: LifecycleCallable):
            self.decorator = decorator
            self.method = method
            self.lifecycle_callable = lifecycle_callable

        def execute(self, instance, environment: Environment):
            self.method.method(instance, *self.lifecycle_callable.args(self.decorator, self.method, environment))

        def __str__(self):
            return f"MethodCall({self.method.method.__name__})"

    # static data

    callables : Dict[object, LifecycleCallable] = {}
    cache : Dict[Type, list[list[AbstractCallableProcessor.MethodCall]]] = {}

    # static methods

    @classmethod
    def register(cls, callable: LifecycleCallable):
        AbstractCallableProcessor.callables[callable.decorator] = callable

    @classmethod
    def compute_callables(cls, type: Type) -> list[list[AbstractCallableProcessor.MethodCall]]:
        descriptor = TypeDescriptor.for_type(type)

        result = [[], [], [], []]  # per lifecycle

        for method in descriptor.get_methods():
            for decorator in method.decorators:
                callable = AbstractCallableProcessor.callables.get(decorator.decorator)
                if callable is not None:  # any callable for this decorator?
                    result[callable.lifecycle.value].append(
                        AbstractCallableProcessor.MethodCall(method, decorator, callable))

        # sort according to order

        result[0].sort(key=lambda call: call.lifecycle_callable.order)
        result[1].sort(key=lambda call: call.lifecycle_callable.order)
        result[2].sort(key=lambda call: call.lifecycle_callable.order)
        result[3].sort(key=lambda call: call.lifecycle_callable.order)

        # done

        return result

    @classmethod
    def callables_for(cls, type: Type) -> list[list[AbstractCallableProcessor.MethodCall]]:
        callables = AbstractCallableProcessor.cache.get(type, None)
        if callables is None:
            callables = AbstractCallableProcessor.compute_callables(type)
            AbstractCallableProcessor.cache[type] = callables

        return callables

    # constructor

    def __init__(self, lifecycle: Lifecycle):
        super().__init__()

        self.lifecycle = lifecycle

    # implement

    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        if lifecycle is self.lifecycle:
            callables = self.callables_for(type(instance))

            for callable in callables[lifecycle.value]:
                callable.execute(instance, environment)

@injectable()
@order(1)
class OnInjectCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_INJECT)

@injectable()
@order(2)
class OnInitCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_INIT)

@injectable()
@order(3)
class OnRunningCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_RUNNING)

@injectable()
@order(4)
class OnDestroyCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_DESTROY)

# the callables

@injectable()
@order(1000)
class OnInitLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_init, Lifecycle.ON_INIT)

@injectable()
@order(1001)
class OnDestroyLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_destroy, Lifecycle.ON_DESTROY)

@injectable()
@order(1002)
class OnRunningLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_running, Lifecycle.ON_RUNNING)

@injectable()
@order(9)
class EnvironmentAwareLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(inject_environment, Lifecycle.ON_INJECT)

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [environment]

@injectable()
@order(10)
class InjectLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(inject, Lifecycle.ON_INJECT)

    # override

    def args(self, decorator: DecoratorDescriptor,  method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [environment.get(type) for type in method.param_types]

def scope(name: str):
    def decorator(cls):
        Scopes.register(cls, name)

        Decorators.add(cls, scope)
        # Decorators.add(cls, injectable)

        Providers.register(ClassInstanceProvider(cls, eager=True, scope="request"))

        return cls

    return decorator

@scope("request")
class RequestScope(Scope):
    # properties

    __slots__ = [
    ]

    # public

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        return provider.create(environment, *arg_provider())

@scope("singleton")
class SingletonScope(Scope):
    # properties

    __slots__ = [
        "value",
        "lock"
    ]

    # constructor

    def __init__(self):
        super().__init__()

        self.value = None
        self.lock = threading.RLock()

    # override

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        if self.value is None:
            with self.lock:
                if self.value is None:
                    self.value = provider.create(environment, *arg_provider())

        return self.value
    
@scope("thread")
class ThreadScope(Scope):
    __slots__ = [
        "_local"
    ]

    # constructor

    def __init__(self):
        super().__init__()
        self._local = threading.local()

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[], list]):
        if not hasattr(self._local, "value"):
            self._local.value = provider.create(environment, *arg_provider())
        return self._local.value

# internal class that is required to import technical instance providers

@environment()
class BootEnvironment:
    # class

    environment = None

    @classmethod
    def get_instance(cls):
        if BootEnvironment.environment is None:
            BootEnvironment.environment = Environment(BootEnvironment)

        return BootEnvironment.environment

    # properties

    __slots__ = []

    # constructor

    def __init__(self):
        pass



"""
Configuration value handling
"""
from .configuration import ConfigurationManager, ConfigurationSource, EnvConfigurationSource, value

__all__ = [
    "ConfigurationManager",
    "ConfigurationSource",
    "EnvConfigurationSource",
    "value"
]
